#!/usr/bin/env python3

# Copyright (C) 2014 Red Hat
#
# This file is part of wikitcms.
#
# wikitcms is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

"""relval manipulates release validation events."""

import argparse
import getpass
import logging
import os
import shutil
import sys
import tempfile
from collections import defaultdict, OrderedDict
from operator import attrgetter
import json

import relval.user_stats as uss
import relval.testcase_stats as tcs
import relval.report_results as rrs
import relval.size_check as rsc

import wikitcms.helpers as hl
import wikitcms.release as rl
import wikitcms.result as rs
import wikitcms.wiki as wk
from mwclient import errors as mwe

logger = logging.getLogger(__name__)

## UTILITY FUNCTIONS ##

def comment_check(string):
    """Just a tiny wrapper to raise the right kind of exception."""
    try:
        return rrs.comment_string(string)
    except ValueError as err:
        raise argparse.ArgumentTypeError(err)

def guess_release(site):
    """Guess the release to operate on: uses the 'next' release so far
    as the wiki is concerned.
    """
    release = int(site.expandtemplates("{{FedoraVersionNumber|next}}"))
    print("Guessing release: {0}".format(str(release)))
    return release

def setup_site(args):
    """Access and, if login is True, log in to either the Fedora wiki
    (test=False) or the Fedora staging wiki (test=anything else, e.g.
    True). Returns a wikitcms Wiki instance.
    """
    if not args.test:
        site = wk.Wiki(('https', 'fedoraproject.org'), '/w/')
    else:
        site = wk.Wiki(('https', 'stg.fedoraproject.org'), '/w/')
    if args.login is True:
        site.login()
    return site

def parse_args():
    """Parse command-line arguments, set up the site instance, and
    call the appropriate sub-command method.
    """
    # Help texts shared by multiple sub-commands.
    release_help = "The Fedora release to operate on (22, 23, 24...)"
    milestone_help_specific = (
        "The milestone to operate on (Alpha, Beta, Final, RC)")
    milestone_help = (
        "The milestone to operate on (Alpha, Beta, Final, RC, Branched, "
        "Rawhide). Must be specified for non-nightly; for a nightly event, "
        "will be guessed if not specified")
    compose_help = "The compose to operate on (1.1, 20160314.n.0)"
    username_help = "DEPRECATED: your wiki (FAS) username. No longer used"
    force_help = "Force creation even if page(s) already exist"
    test_help = "Operate on the staging wiki (for testing)"
    filter_help = ("Keyword(s) to filter pages that will be processed. Can "
                   "be specified more than once, all pages matching ANY "
                   "specified keyword will be included")

    # See the argparse instructions for all of this. Use of 'metavar'
    # avoids the entire ranges of valid values being shown in the help
    parser = argparse.ArgumentParser(
        description=(
            "Tool for various tasks related to Fedora release validation "
            "testing. 'compose' can create result pages for TC/RC and nightly "
            "validation test events. 'user-stats' and 'testcase-stats' can "
            "output statistics related to validation testing. 'report-results'"
            " lets you report test results. See each command's help for more "
            "details."
        ),
        epilog=(
            "For commands which require logging in to the wiki, on first use, "
            "a browser window will open and walk you through the Fedora auth "
            "process. An authentication token will be created to authenticate "
            "you for subsequent uses. After a while, the token will expire, "
            "and you will go through the browser authentication process again."
        )
    )
    parser.add_argument(
        '-l', '--loglevel', help="The level of log messages to show",
        choices=('debug', 'info', 'warning', 'error', 'critical'),
        default='info')
    # python 3 issue workaround, see:
    # https://stackoverflow.com/questions/23349349
    # http://bugs.python.org/issue16308
    subparsers = parser.add_subparsers(dest='subcommand')
    subparsers.required = True

    parser_compose = subparsers.add_parser(
        'compose', description="Create Fedora release validation event result "
        "pages. You must specify EITHER --release, --milestone and --compose "
        "OR --cid. You can create an event before the compose has completed, "
        "but you may not be able to use --cid, and the download page will "
        "not be created, you must remember to create that after the compose "
        "completes. This command requires login; see main help information "
        "about authentication.")
    parser_compose.add_argument(
        '-r', '--release', help=release_help, type=int, metavar="24-99",
        choices=list(range(24, 100)))
    parser_compose.add_argument(
        '-m', '--milestone', help=milestone_help,
        choices=['Beta', 'Final', 'RC', 'Branched', 'Rawhide'])
    parser_compose.add_argument(
        '-c', '--compose', help=compose_help, metavar="1.1 or "
        "20160314.n.0")
    parser_compose.add_argument(
        '--modular', help="Operate on modular compose", action='store_true')
    parser_compose.add_argument(
        '-i', '--cid', help="Compose ID of a Pungi 4 compose to create a "
        "validation event for. If this is a 'production' compose - i.e. "
        "a candidate compose, not a nightly compose - it must be "
        "completed.", metavar="24-20160314.n.0")
    parser_compose.add_argument(
        '-y', '--testtype', help="The test type to generate a page for")
    parser_compose.add_argument(
        '-u', '--username', help=username_help)
    parser_compose.add_argument(
        '-f', '--force', help=force_help, action='store_true')
    parser_compose.add_argument(
        '-t', '--test', help=test_help, action='store_true')
    parser_compose.add_argument(
        '-n', '--no-current', help="Do not update Test Results:Current "
        "redirect pages", action='store_true')
    parser_compose.add_argument(
        '-w', '--download-only', help="Only write/update the Download page, "
        "do not touch any other page", action='store_true')
    parser_compose.set_defaults(func=create_compose, login=True, guessrel=True)

    parser_userstats = subparsers.add_parser(
        'user-stats', description="Output statistics regarding how many tests "
        "were run and bugs reported by each individual tester for a set of "
        "test events.")
    parser_userstats.add_argument(
        '-r', '--release', help=release_help, type=int, metavar="12-99",
        choices=list(range(12, 100)))
    parser_userstats.add_argument(
        '-m', '--milestone', help=milestone_help_specific,
        choices=['Alpha', 'Beta', 'Final', 'RC'])
    parser_userstats.add_argument(
        '--modular', help="Operate on modular composes", action='store_true')
    parser_userstats.add_argument(
        '-t', '--test', help=test_help, action='store_true')
    parser_userstats.add_argument(
        '-f', '--filter', metavar='KEYWORD', help=filter_help, action='append')
    parser_userstats.add_argument(
        '-s', '--since', help='Only include pages created on or after this '
        'date', metavar='YYYYMMDD')
    parser_userstats.add_argument(
        '-u', '--until', help='Only include pages created before or on this '
        'date', metavar='YYYYMMDD')
    parser_userstats.add_argument(
        '-b', '--bot', action='store_true',
        help="Include results from automated tests ('bot results')")
    parser_userstats.set_defaults(func=user_stats, login=False, guessrel=True)

    parser_tcstats = subparsers.add_parser(
        'testcase-stats', description="Output a set of HTML pages containing "
        "tables providing an overview of test coverage across a series of "
        "test events.")
    parser_tcstats.add_argument(
        '-r', '--release', help=release_help, type=int, metavar="12-99",
        choices=list(range(12, 100)))
    parser_tcstats.add_argument(
        '-m', '--milestone', help=milestone_help_specific,
        choices=['Alpha', 'Beta', 'Final', 'RC'])
    parser_tcstats.add_argument(
        '--modular', help="Operate on modular composes", action='store_true')
    parser_tcstats.add_argument(
        '-o', '--out', help="The directory to write the pages to. WILL BE "
        "OVERWRITTEN if it already exists", metavar='/some/path')
    parser_tcstats.add_argument(
        '-t', '--test', help=test_help, action='store_true')
    parser_tcstats.add_argument(
        '-f', '--filter', metavar='KEYWORD', help=filter_help, action='append')
    parser_tcstats.add_argument(
        '-n', '--nopostproc', help="Do not post-process the results to try "
        "and clean moved test cases and section order", action='store_true')
    parser_tcstats.set_defaults(
        func=testcase_stats, login=False, guessrel=True)

    parser_result = subparsers.add_parser(
        'report-results', description="Submit validation test results "
        "for an event. This command requires login; see main help "
        "information about authentication.")
    parser_result.add_argument(
        '-u', '--username', help=username_help)
    parser_result.add_argument(
        '-r', '--release', help=release_help, type=int, metavar="12-99",
        choices=list(range(12, 100)))
    parser_result.add_argument(
        '-m', '--milestone', help=milestone_help,
        choices=['Beta', 'Final', 'RC', 'Branched', 'Rawhide'])
    parser_result.add_argument(
        '-c', '--compose', help=compose_help, metavar="1.1 or 20160314.n.0")
    parser_result.add_argument(
        '--modular', help="Operate on modular compose", action='store_true')
    parser_result.add_argument(
        '-y', '--testtype', help="The test type to report a result for (Base, "
        "Cloud...)")
    parser_result.add_argument(
        '-s', '--section', help="The page section to report the result in")
    parser_result.add_argument(
        '-t', '--test', help=test_help, action='store_true')
    parser_result.set_defaults(func=report_results, login=True, guessrel=False)

    parser_sizecheck = subparsers.add_parser(
        'size-check', description="Check image sizes for a compose and "
        "report the results.")
    parser_sizecheck.add_argument(
        '-u', '--username', help=username_help)
    parser_sizecheck.add_argument(
        '-r', '--release', help="The Fedora release to check", type=int,
        choices=list(range(1, 100)), metavar="1-99")
    parser_sizecheck.add_argument(
        '-m', '--milestone', help="A milestone to check (e.g. Beta)",
        choices=['Beta', 'Final', 'RC', 'Branched', 'Rawhide'])
    parser_sizecheck.add_argument(
        '-c', '--compose', help=compose_help, metavar="1.1 or 20160314.n.0")
    parser_sizecheck.add_argument(
        '-i', '--cid', help="Compose ID of a Pungi 4 compose to check",
        metavar="24-20160314.n.0")
    parser_sizecheck.add_argument(
        '--modular', help="Operate on modular compose", action='store_true')
    parser_sizecheck.add_argument(
        '-t', '--test', help=test_help, action='store_true')
    parser_sizecheck.set_defaults(func=size_check, login=True, guessrel=False)

    args = parser.parse_args()

    site = setup_site(args)
    # guessrel is kinda ugly, but some sub-functions don't want release
    # guessing
    if args.guessrel and not args.release and not getattr(args, 'url', ''):
        args.release = guess_release(site)
    if hasattr(args, 'username') and args.username:
        print("WARNING: the '--username' argument no longer has any effect "
              "and may be removed in a future release. See the README for "
              "more information on the newer auth system from 2018 onwards.")

    return (args, site)

## SUB-COMMAND METHODS ##

def create_compose(args, site):
    """Create the result pages for a given event (and, optionally,
    test type). Uses parameters passed in from the command line.
    """
    if not args.cid and not (args.release and args.milestone and
                             args.compose):
        sys.exit("You must specify --cid or --milestone and --compose!")
    if args.cid and (args.release or args.milestone or args.compose):
        sys.exit("You specified both --cid and at least one of --release, "
                 "--milestone and --compose. This is ambiguous and not "
                 "allowed. Please specify EITHER --cid OR --release, "
                 "--milestone and --compose.")
    try:
        event = site.get_validation_event(args.release, args.milestone,
                                          args.compose, args.cid, modular=args.modular)
    except ValueError as err:
        sys.exit(err)

    if args.download_only:
        print("Only handling download page!")
        try:
            # handling force is...annoying. createonly must be 'None'
            # to force writing (not False)
            createonly = True
            if args.force:
                createonly = None
            event.download_page.write(createonly=createonly)
            return
        except mwe.APIError as err:
            sys.exit(err)

    if args.testtype:
        testtypes = [args.testtype]
    else:
        testtypes = None
    current = not args.no_current
    try:
        event.create(
            testtypes=testtypes, force=args.force, current=current)
    except mwe.APIError:
        print("Unhandled error writing pages!")
        raise
    except ValueError as err:
        sys.exit(err)

def user_stats(args, site):
    """Print HTML-formatted statistics on the number of results
    reported by each tester for a given set of result pages.
    """
    release = rl.Release(args.release, site, modular=args.modular)
    pages = release.milestone_pages(args.milestone)
    if args.filter:
        pages = [p for p in pages for f in args.filter if f in p.name]
    if args.since:
        pages = [p for p in pages if int(p.creation_date) >= int(args.since)]
    if args.until:
        pages = [p for p in pages if int(p.creation_date) <= int(args.until)]
    text = []
    # dump the results tables from each page into a single big string
    for page in pages:
        sys.stderr.write("Processing page: {}\n".format(page.name))
        text.append(page.results_wikitext)
    text = '\n'.join(text)

    # find results with the statuses we care about that are not transferred
    # from previous runs
    results = rs.find_results(text, statuses=['pass', 'fail', 'warn'],
                              transferred=False, bot=args.bot)

    # make two defaultdicts for number of reports and cited bugs per unique
    # user. defaultdict(int) sets the value when adding a key to int(0).
    reports = defaultdict(int)
    bugs = defaultdict(set)
    for result in results:
        if result.user:
            reports[result.user.lower()] += 1
        if result.bugs:
            bugs[result.user].update(result.bugs)

    uss.print_stats(args.release, reports, bugs)

def testcase_stats(args, site):
    """Generate HTML-formatted statistics on the test coverage for
    a given release - produces summary pages for each test type
    containing tables with a row for each test indicating the level
    of coverage at each compose, and detail pages for each test
    with detailed statistics on the test results and bugs for that
    test for each compose.
    """
    outdir = tempfile.mkdtemp(prefix='tc_report_')
    release = rl.Release(args.release, site, modular=args.modular)
    relpages = release.milestone_pages(args.milestone)
    if args.filter:
        relpages = [p for p in relpages for f in args.filter if f in
                    p.name]

    allpages = defaultdict(list)
    # for sorting purposes, we start out with the record of composes
    # as a dict, key is the compose version, value is its date
    allcomposes = dict()
    for page in relpages:
        allpages[page.testtype].append(page)
        if page.shortver not in allcomposes.keys():
            allcomposes[page.shortver] = page.creation_date

    # turn the composes dict into a list and sort it by creation date,
    # with wikitcms sort as a tiebreaker (viz 23 Final RC7 vs. RC9)
    allcomposes = list(allcomposes.items())
    allcomposes.sort(key=lambda x: (x[1], hl.fedora_release_sort(x[0])))
    allcomposes = [comp[0] for comp in allcomposes]

    # for each test type, sort the pages into the correct Fedora release
    # order, then extract the result rows from each page.
    #
    # 'tests' is an ordered dict of Test() objects, keyed with a tuple
    # identifying 'unique tests': the test 'name', its testcase page name,
    # and the name of the page section in which it appears. It is ordered
    # so that tests appear in the order they are encountered - when the
    # HTML output is rendered, the ordering of tests will be more or less
    # the same as on the actual result wiki pages. tcs.post_process()
    # cleans this up some more once all result pages have been parsed. See
    # its docs for details.
    #
    # A Test() object is generated for each result row from each result
    # page for the test type, and added to the tests dict based on those
    # three attributes (as the second field of a tuple, the first field
    # identifying the compose from which it comes). Result rows for which
    # all three attributes match are grouped together - they're considered
    # to be results for the same test in different composes.
    jsondata = {'allcomposes': allcomposes, 'stats': dict()}
    for index, (testtype, pages) in enumerate(allpages.items()):
        print('Processing type [{0}/{1}]: {2}'.format(index + 1,
            len(allpages), testtype))
        tests = OrderedDict()
        # re-use the allcomposes sort to sort the pages.
        pages.sort(key=lambda x: allcomposes.index(x.shortver))
        for index, page in enumerate(pages):
            print("Processing page [{0}/{1}]: {2}".format(index + 1,
                len(pages), page.name))
            resultrows = page.get_resultrows(
                statuses=['pass', 'warn', 'fail'], transferred=False)
            for row in resultrows:
                tests.setdefault((row.name, row.testcase, row.section),
                                 tcs.Test())
                tests[(row.name, row.testcase, row.section)].update(
                    page.shortver, row)

        if args.nopostproc:
            print("nopostproc passed! Skipping post-process.")
        else:
            print("Post-processing results...")
            tests = tcs.post_process(tests, allcomposes)
        tcs.print_results_html(testtype, pages, allcomposes, tests, outdir)
        jsondata['stats'][testtype] = tcs.get_results_data(allcomposes, tests)

    tcs.print_summary_html(outdir, allpages)
    with open(os.path.join(outdir, 'data.json'), 'w') as outfile:
        json.dump(jsondata, outfile, sort_keys=True, indent=4)

    if args.out:
        try:
            shutil.rmtree(args.out)
        except FileNotFoundError:
            pass
        except PermissionError:
            logger.warning("Could not clean output dir %s!", args.out)
        try:
            shutil.move(outdir, args.out)
            print("All output moved to: {0}".format(args.out))
            print("Index page at: {0}".format(os.path.join(args.out,
                                                           "index.html")))
            print("JSON data at: {0}".format(os.path.join(args.out,
                                                           "data.json")))
        except (PermissionError, FileExistsError):
            print("Could not write to: {0}! "
                  "Output is at: {1}".format(args.out, outdir))
    else:
        print("HTML Output: " + os.path.join(outdir, "index.html"))
        print("JSON Output: " + os.path.join(outdir, "data.json"))
        print(outdir)

def report_results(args, site):
    """Report some validation testing results."""
    rrs.report_results(site, args.release, args.milestone, args.compose,args.testtype,
                       modular=args.modular)

def size_check(args, site):
    """Check sizes of images for event's compose and report results."""
    try:
        event = site.get_validation_event(
            args.release, args.milestone, args.compose, args.cid, modular=args.modular)
        ffrel = event.ff_release
        print("Checking sizes for event: "
              "{0}, compose ID: {1}".format(event.version, ffrel.cid))
    except ValueError:
        sys.exit("Could not find validation test event!")
    results = rsc.size_check(ffrel)
    if not results:
        sys.exit("No images found for event!")
    reslist = list()
    for (env, result) in list(results.items()):
        if result == 'pass':
            restup = wk.ResTuple(
                'Installation', args.release, args.milestone, args.compose,
                'ISO_Size', '', '', env, result, args.username, '', '',
                True, args.cid, args.modular)
        else:
            # Fail - construct a comment with details
            comms = []
            for fail in result:
                desc = ' '.join((fail[0]['subvariant'], fail[0]['type'], fail[0]['arch']))
                comms.append("{0}, size {1}, max {2}".format(desc, fail[1],
                                                             fail[2]))
            comment = "<ref>{0}</ref>".format('</ref><ref>'.join(comms))
            # Mark this as a bot result as it's an automated test.
            restup = wk.ResTuple(
                'Installation', args.release, args.milestone, args.compose,
                'ISO_Size', '', '', env, 'fail', args.username, '',
                comment, True, args.cid, args.modular)
        reslist.append(restup)
    (insuffs, dupes) = site.report_validation_results(reslist)
    if dupes:
        for dupe in dupes:
            print("This user already reported a result for {0}! Will not "
                  "file duplicate report.".format(dupe.env))
    if insuffs and len(insuffs) == len(results):
        sys.exit("Could not find result page or test case!")
    if insuffs:
        for insuff in insuffs:
            print("Could not find environment {0}!".format(insuff.env))
    sys.exit(0)

## MAIN LOOP BITS ##

def main():
    '''Main loop'''
    try:
        args, site = parse_args()
        loglevel = getattr(
            logging, args.loglevel.upper(), logging.INFO)
        logging.basicConfig(level=loglevel)
        args.func(args, site)
    except KeyboardInterrupt:
        print('Interrupted, exiting...')
        sys.exit(1)

if __name__ == '__main__':
    main()
