# Copyright (C) 2014 Red Hat
#
# This file is part of relval.
#
# wikitcms is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
# Adam Williamson <awilliam@redhat.com>
#
"""This file contains functions that help implement the report-results
sub-command of relval, which is a simple utility to report release
validation results.
"""

import sys
import re

import fedfind.helpers
import wikitcms.result
from productmd.composeinfo import get_date_type_respin

# these are some values we cache across multiple runs of the main
# function.
RESPAGES = None
PAGE = None
SECTIONS = None

def comment_string(string, maxlen=250):
    """Take 'string' and wrap it in <ref> </ref> if it's not the empty
    string. Raise a ValueError if it's longer than maxlen.
    """
    string = string.strip()
    if maxlen and len(string) > maxlen:
        err = "Comment is too long: {0} characters, max is {1}.".format(
            str(len(string)), str(maxlen))
        raise ValueError(err)
    # Don't produce an empty <ref></ref> if there's no comment
    if not string:
        return string
    return "<ref>{0}</ref>".format(string)

def check_nightly(string):
    """Check if a string looks like a valid nightly identifier."""
    try:
        (date, typ, respin) = get_date_type_respin(string)
        if typ == 'nightly':
            return True
    except ValueError:
        return False

# mini functions for prompting the user for choices

def get_response(prompt, helptext, accepted=None, accfunc=None):
    """Get a response to a question. Prompt is the initial text,
    helptext is help text shown on invalid response, accepted is an
    iterable of valid responses (case-insensitive comparison and
    return). accfunc is a function to call on the response to
    validate it. Either accepted or accfunc must be given.
    """
    if not accepted and not accfunc:
        raise ValueError(
            "get_response: Must specify accepted or accfunc!")
    def check(resp, accepted, accfunc):
        """'accepted' makes things a mite tricky."""
        if accfunc:
            return accfunc(resp.lower())
        elif accepted:
            return resp.lower() in accepted

    resp = input(prompt)
    while not check(resp, accepted, accfunc):
        print("Invalid response. {0}".format(helptext))
        resp = input(prompt)
    return resp.lower()

def get_confirm():
    """Just ask "yes" or "no" and return 'y' or 'n'."""
    return get_response("Confirm y/n? ", helptext="Answer y or n",
                        accepted=('y', 'n'))

def get_response_list(prompt, items):
    """Given a question and an iterable of strings, number the strings
    and ask the user which they want. Returns the int for the user's
    selection, minus 1 (because you probably want to number from 1 for
    humans but use the result as a list index).
    """
    lines = []
    for (num, item) in enumerate(items, 1):
        lines.append("{0})  {1}".format(num, item))
    lines.append(prompt)
    prompt = '\n'.join(lines)
    helptext = "Pick a number between 1 and {0}".format(str(len(items)))
    accepted = [str(num) for num in range(1, len(items)+1)]
    return int(get_response(prompt, helptext, accepted=accepted))-1

def get_result(username):
    """Get a wikitcms Result instance by asking the user for all the
    required information.
    """
    prompt = "Would you like to report a (p)ass, (f)ail or (w)arn? "
    helptext = "Please choose (p)ass, (f)ail or (w)arn."
    status = get_response(prompt, helptext, accepted=('p', 'f', 'w'))
    for (let, full) in (('p', 'pass'), ('f', 'fail'), ('w', 'warn')):
        if status == let:
            status = full

    bugs = input("Enter the IDs of any bugs associated with this report, "
                 "separated by commas. ")
    if bugs == '':
        bugs = None
    else:
        white = re.compile(r'\s')
        bugs = white.sub('', bugs)
        bugs = bugs.split(',')
        if bugs == '':
            bugs = None
    comment = ''
    comm = input("Enter a comment if you like. If not, just hit Enter.\n"
                 "Maximum length is 250 characters.\nPlease don't leave a "
                 "comment unless it's really necessary.\nComment: ")
    while comm:
        try:
            comment = comment_string(comm, 250)
            comm = ''
        except ValueError:
            comm = input("Comment was too long: {0} characters. Please make "
                         "it 250 characters or fewer.\n"
                         "Comment: ".format(len(comm.strip())))
    return wikitcms.result.Result(status, username, bugs, comment)


def report_results(wiki, release=None, milestone='', compose='',
                   testtype=None, section=None, test=None, auto=True,
                   checkver=True, modular=None):
    """Main function, find a test instance to report a result for and
    create a result, then report it.
    """
    try:
        username = wiki.username.lower()
    except:
        sys.exit("ERROR: Cannot find wiki username! Not logged in?")
    # these are global so they can be cached between runs
    global RESPAGES
    global PAGE
    global SECTIONS
    version_set = False
    version_type = ''
    dist = "Fedora"
    if modular:
        dist = "Fedora Modular"

    # this means 'try and detect the current compose, unless we already
    # have one fully specified or we've been told not to'
    if not (release and compose) and auto is True:
        if release or compose:
            print("When specifying the compose version by parameters, you "
                  "must pass at least a release and a compose. Your "
                  "version-related parameters will not be used.")
        print("Detecting current compose version...")
        if modular:
            curr = wiki.current_modular_compose
        else:
            curr = wiki.current_compose
        release = int(curr['release'])
        milestone = curr['milestone']
        # this is a handy way to get whichever is set, and if for some
        # crazy reason both are set it'll prefer 'T/RCx' to 'YYYYMMDD',
        # which we probably want.
        compose = max(curr['compose'], curr['date'])

    if release and compose:
        version_set = True

    if modular is None:
        prompt = "Are you reporting results for a Modular compose? "
        resp = get_response(prompt, helptext="Answer y or n",
                        accepted=('y', 'n'))
        if resp == 'y':
            modular = True
        else:
            modular = False

    validrels = [str(num) for num in range(23, 100)]
    if str(release) not in validrels:
        prompt = "Which Fedora release would you like to report a result for? "
        helptext = "Please input a valid Fedora release number."
        release = int(get_response(prompt, helptext, accepted=validrels))

    while not version_set:
        if not version_type:
            prompt = ("Would you like to report a result for a milestone "
                      "(c)ompose or a (n)ightly compose? ")
            helptext = "Please choose (c)ompose or (n)ightly."
            version_type = get_response(prompt, helptext, accepted=('c', 'n'))

        if version_type == 'c':
            prompt = ("(B)eta\n(R)C\nWhich milestone would "
                      "you like to report a milestone for? ")
            helptext = "Please choose (B)eta, or (R)C."
            milestone = get_response(prompt, helptext, accepted=('b', 'r'))
            for (let, full) in (('b', 'Beta'), ('r', 'RC')):
                if milestone == let:
                    milestone = full
            if not compose:
                prompt = "Which compose would you like to report a result for? "
                helptext = "Examples: 1.0, 1.1, 2.1"
                compose_patt = re.compile(r'\d+\.\d+')
                compose = get_response(
                    prompt, helptext, accfunc=compose_patt.match).upper()
            version_set = True
        else:
            if not compose:
                prompt = "What nightly would you like to report a result for? "
                helptext = "Nightly must be YYYYMMDD.n.N (e.g. 20160314.n.0)"
                compose = get_response(prompt, helptext, accfunc=check_nightly)
            version_set = True

    if milestone:
        dispver = "{0} {1} {2} {3}".format(dist, release, milestone, compose)
    else:
        dispver = "{0} {1} {2}".format(dist, release, compose)
    if checkver:
        print("Will report results for compose: {0}".format(dispver))
        confirm = get_confirm()
        if confirm != 'y':
            report_results(wiki, auto=False)

    if not RESPAGES:
        try:
            event = wiki.get_validation_event(
                release=release, milestone=milestone, compose=compose, modular=modular)
        except ValueError as err:
            print(err)
            report_results(wiki)

        RESPAGES = event.result_pages
        if not RESPAGES:
            print("Error! You probably chose a compose that doesn't exist.")
            report_results(wiki)

    if PAGE is None:
        if testtype:
            try:
                PAGE = [pag for pag in RESPAGES if pag.testtype == testtype][0]
            except IndexError:
                print("Error! Specified test type does not exist.")
        if not PAGE:
            prompt = "Which test type would you like to report a result for? "
            items = [page.testtype for page in RESPAGES]
            PAGE = RESPAGES[get_response_list(prompt, items)]

    tests = PAGE.get_resultrows(statuses=['pass', 'warn', 'fail', 'inprogress'])

    if not section:
        try:
            # This serves two purposes: if we somehow wound up with a dead
            # page, we can try again...
            if len(SECTIONS) == 0:
                print("Error! No results found for this page.")
                report_results(wiki)
        except TypeError:
            # ...and if we hit an exception, it means we need to get SECTIONS.
            SECTIONS = PAGE.results_sections
            # Filter out sections with no tests in them. wikitcms should
            # probably do this but it's actually not trivial for $REASONS
            testsecs = {t.secid for t in tests}
            SECTIONS = [s for s in SECTIONS if s['index'] in testsecs]
        if len(SECTIONS) == 1:
            section = SECTIONS[0]
        else:
            prompt = "Which section would you like to report a result in? "
            tag_patt = re.compile('<.*?>')
            items = [tag_patt.sub('', section['line']) for section in SECTIONS]
            section = SECTIONS[get_response_list(prompt, items)]

    # Find the tests for the section the user chose.
    tests = [t for t in tests if t.secid == section['index']]
    if len(tests) == 1:
        test = tests[0]
    else:
        prompt = "Which test would you like to report a result for? "
        items = []
        for test in tests:
            if test.name != test.testcase:
                items.append("{0} {1}".format(test.testcase, test.name))
            else:
                items.append(test.testcase)
        test = tests[get_response_list(prompt, items)]

    envs = list(test.results.keys())
    if len(envs) == 1:
        env = envs[0]
    elif len(envs) > 1:
        prompt = "Which environment would you like to report a result for? "
        env = envs[get_response_list(prompt, envs)]

    results = test.results[env]
    if results:
        print("Current results for this test:")
        for result in results:
            print(str(result))
        myres = [r for r in results if r.user == username]
        if myres:
            print("You have already reported result(s) for this test! Do you "
                  "still want to report this result?")
            confirm = get_confirm()
            if confirm == 'y':
                pass
            else:
                report_results(wiki, release, milestone, compose, modular=modular)

    res = get_result(username)
    print("Will report {0} for test {1} in section "
          "{2} of {3}".format(res, test.testcase, section['line'], PAGE.name))
    try:
        PAGE.add_result(res, test, env)
        print("Succeeded!")
    except:
        raise

    # we're done, what next?
    prompt = ("Report another result for this (s)ection, or (p)age, "
              "or this (c)ompose, or (q)uit? ")
    helptext = "Please choose (s)ection, (p)age, (c)ompose, or (q)uit."
    cont = get_response(prompt, helptext, ('s', 'p', 'c', 'q'))
    if cont == 'q':
        sys.exit()
    else:
        test = None
    if cont == 'p' or cont == 'c':
        section = None
    if cont == 'c':
        testtype = None
        PAGE = None
        SECTIONS = None
    report_results(wiki, release, milestone, compose, testtype,
                   section, checkver=False, modular=modular)
